import re


class RancherWorkload:
    def __init__(self, value):
        parts = value.split(':')
        if len(parts) == 3:
            self.type, self.namespace, self.name = parts
        elif len(parts) == 2:
            self.type = 'deployment'
            self.namespace, self.name = parts
        else:
            raise Exception("Invalid workload")
        self.id = ":".join([self.type, self.namespace, self.name])

    def __repr__(self):
        return self.id


class RancherPod:
    def __init__(self, values):
        self.workload = RancherWorkload(values["workloadId"])
        self.id = values["id"]
        self.name = self.id.split(":")[1]

    def __repr__(self):
        return self.id


class DockerImage:
    def __init__(self, value, partial=False):
        rex = re.compile(r"(?P<image>.*):?(?P<tag>[\w\.-]*)")
        m = rex.match(value)
        self.image, self.tag = m.groups()
        if self.tag and not self.image:
            raise Exception("Tag needs and image")
        if self.tag:
            self.id = f"{self.image}:{self.tag}"
        elif self.image:
            self.id = f"{self.image}"
        else:
            raise ValueError(value)

        #     self.id = self.account

    def __repr__(self):
        return self.id



import re
import ssl
import time
from functools import wraps
from json import JSONDecodeError
from urllib.parse import urlparse

import _thread
import websocket
from click import UsageError
from requests import request
from requests.exceptions import SSLError

from lazo.objects import RancherPod
from lazo.types import RancherWorkload

from .exceptions import (HttpError, InvalidCredentials, InvalidName,
                         ObjectNotFound, ServerConnectionError, ServerSSLError,)
from .out import echo, error, fail
from .utils import jprint

success_codes = (200, 201, 204)


def on_message(ws, message):
    print(111, message)


def on_error(ws, error):
    print(222, error)


def on_close(ws):
    print(333, "### closed ###")


def on_open(ws):
    def run(*args):
        time.sleep(1)
        ws.close()
        print("thread terminating...")

    _thread.start_new_thread(run, ())


class HttpClient:
    def __init__(self, base_url, *, verify=True, debug=False, auth=None, **kwargs):
        o = urlparse(base_url)
        self.scheme = o.scheme or 'http'
        self.port = o.port or {"http":80, "https":443}[self.scheme]
        self.scheme = o.scheme
        self.host = o.hostname
        self.path = o.path

        self.base_url = base_url
        self.verify = verify
        self.debug = debug
        self.history = []
        self.auth = auth

    def ping(self) -> bool:
        self.get('/')
        return self.history[-1].status_code == 200

    def _r(self, cmd, url, *, raw=False, ignore_error=False, **kwargs):
        if not (url.startswith('http') or url.startswith('wss')):
            url = f"{self.base_url}{url}"
        try:
            if self.debug:
                echo(f"{cmd} {url}")
            response = request(cmd, url, auth=self.auth, verify=self.verify, **kwargs)
        except SSLError:
            raise ServerSSLError(url)
        except Exception as e:
            raise ServerConnectionError(url, e)
        self.history.append(response)

        if response.status_code in success_codes:
            if raw:
                return response
            else:
                try:
                    return response.json()
                except JSONDecodeError as e:
                    raise HttpError(url, response, e)
        elif ignore_error:
            return response
        else:
            raise HttpError(url, response)

    def post(self, url, **kwargs):
        return self._r('post', url, **kwargs)

    def get(self, url, **kwargs):
        return self._r('get', url, **kwargs)

    def delete(self, url, **kwargs):
        return self._r('delete', url, **kwargs)

    def put(self, url, *, data, **kwargs):
        return self._r('put', url, json=data, **kwargs)

    def ws(self, where):
        scheme = 'ws' if self.scheme == 'http' else 'wss'
        base = f'{scheme}://{self.host}:{self.port}'
        url = f"{base}{where}"
        if self.debug:
            echo(f"WS {url}")
        from base64 import b64encode
        userAndPass = b64encode(f"{self.auth.username}:{self.auth.password}".encode('utf8')).decode("ascii")
        headers = {'Authorization': 'Basic %s' % userAndPass}
        ws = websocket.create_connection(url,
                                         sslopt={"cert_reqs": ssl.CERT_NONE},
                                         header=headers)
        assert ws.connected
        data = []
        while ws.connected:
            rec = ws.recv()
            if rec != b'\x01':
                if isinstance(rec, str):
                    data.append(rec[1:])
                else:
                    data.append(rec[1:].decode('utf8'))

        if not data:
            raise Exception('No response')

        return "".join(data)


class RancherClient(HttpClient):
    def __init__(self, base_url, **kwargs):
        self.use_names = kwargs.pop('use_names', False)
        super().__init__(base_url, **kwargs)
        self._cluster = None
        self._project = None

    def __repr__(self):
        return f"<RancherClient {self.base_url}>"

    @property
    def cluster(self):
        return self._cluster

    @cluster.setter
    def cluster(self, cluster_id):
        if self.use_names:
            cluster_id = self._get_cluster_id_by_name(cluster_id)
        self._cluster = cluster_id

    @property
    def project(self):
        return self._project

    @project.setter
    def project(self, name_or_id):
        if isinstance(name_or_id, (list, tuple)):
            if name_or_id[0]:
                self._cluster = name_or_id[0]
            name = name_or_id[1]
        else:
            name = name_or_id

        if self.use_names:
            name = self._get_project_id_by_name(name)
        self._project = name

    # pods
    def list_pods(self):
        url = f'/project/{self.cluster}:{self.project}/pods?limit=-1&sort=name'
        res = self.get(url)
        return res['data']

    def get_pod(self, workload: RancherWorkload, elem=1):
        url = f'/project/{self.cluster}:{self.project}/pods?limit=-1&sort=name'
        res = self.get(url)
        for w in res["data"]:
            if w["workloadId"] == workload.id:
                return RancherPod(w)
        raise ObjectNotFound(workload.id)

    # containers
    def list_containers(self):
        url = f'/project/{self.cluster}:{self.project}/containers?limit=-1&sort=name'
        res = self.get(url)
        return res['data']

    def list_clusters(self):
        res = self.get('/clusters')
        return [(e['name'], e['id']) for e in res['data']]

    def list_projects(self):
        response = self.get(f'/clusters/{self.cluster}/projects')
        return [(e['name'], e['id']) for e in response['data']]

    def list_workloads(self):
        response = self.get(f'/projects/{self.cluster}:{self.project}/workloads')
        return [(e['name'], e['id']) for e in response['data']]

    def get_workload(self, name):
        response = self.get(f'/projects/{self.cluster}:{self.project}/workloads/{name}')
        return response

    def upgrade(self, workload, image):
        url = f'/project/{self.cluster}:{self.project}/workloads/{workload.id}'
        response = self.get(url)
        json = response.copy()
        found = set()
        if 'containers' in response:
            for pod in json['containers']:
                found.add(pod['image'])
                pod['image'] = image.id
        return self.put(url, data=json)

    def _get_cluster_id_by_name(self, name):
        res = self.get('/clusters')
        for entry in res['data']:
            if entry['name'] == name:
                return entry['id']
        raise InvalidName(f"Invalid cluster name '{name}'")

    def _get_project_id_by_name(self, name):
        res = self.get(f'/clusters/{self.cluster}/projects')
        for entry in res['data']:
            if entry['name'] == name:
                return entry['id'].split(":")[1]
        raise InvalidName(f"Invalid project name '{name}'")

    def _get_workload_id_by_name(self, project, name):
        response = self.get(f'/projects/{self.cluster}:{self.project}/workloads')
        for workload in response['data']:
            if workload['name'] == name[-1]:
                return workload['id'].split(":")
        raise InvalidName(f"Invalid workload name '{':'.join(name)}'")


class DockerClient(HttpClient):

    def __init__(self, base_url, *, username=None, password=None, **kwargs) -> None:
        super().__init__(base_url, **kwargs)
        self.username = username
        self.password = password

    def __repr__(self):
        return f"<DockerClient {self.base_url}>"

    def exists(self, image):
        try:
            self.get(f'/repositories/{image.image}/tags/{image.tag}/')
            return True
        except HttpError:
            return False

    def login(self):
        url = '/users/login/'
        self.post(url, json={"username": self.username,
                             "password": self.password},
                  ignore_error=True)
        response = self.history[-1]
        if response.status_code == 400:
            raise InvalidCredentials(url, response)
        self.token = response.json()['token']
        return self.token

    def get_tags(self, image, filter='.*', max_pages=None):
        ret = []
        url = f'/repositories/{image.image}/tags/'
        rex = re.compile(filter)
        page = 1
        while url:
            # TODO: remove me
            print(111, "clients.py:272", url)
            response = self.get(url)
            for e in response['results']:
                if rex.search(e['name']):
                    yield e
            if max_pages and page > max_pages:
                break
            url = response['next']
        return sorted(ret)


def handle_lazo_error(func):
    @wraps(func)
    def inner(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except ServerSSLError as e:
            fail(e)
        except UsageError as e:
            error(e)
        except HttpError as e:
            data = e.response.json()
            error(e.url)
            jprint(data)
        except ServerConnectionError as e:
            error(str(e))
            error(str(e.reason))

    return inner



import sys

import click

from ..__cli__ import cli
from ..clients import DockerClient, handle_lazo_error
from ..exceptions import HttpError
from ..out import echo, error, success
from ..params import _docker_options, _global_options, make_option, options
from ..types import Image
from ..utils import jprint, sizeof


@cli.group()
@options(_global_options, _docker_options)
@click.pass_context
def docker(ctx, repository, username, password, stdin, **kwargs):
    if stdin:
        username, password = stdin
    ctx.obj['client'] = DockerClient(repository,
                                     username=username, password=password,
                                     debug=ctx.find_root().command.debug)


@docker.command()
@options(_global_options)
@click.argument("image", type=Image, metavar='IMAGE')
@click.option("--filter", metavar='REGEX', default='.*')
@click.option("--size", is_flag=True)
@click.pass_context
def info(ctx, image, filter, size, **kwargs):
    client = ctx.obj['client']
    try:
        if image.tag:
            response = client.get(f'/repositories/{image.image}/tags/{image.tag}/')
            jprint(response)
        elif image.image:
            for tag in client.get_tags(image, filter):
                if size:
                    echo(f"{tag['name']:<30} {sizeof(tag['full_size'])}")
                else:
                    echo(tag['name'])

        elif image.account:
            response = client.get(f'/repositories/{image.account}/')
            for e in response['results']:
                echo(e['name'])
    except HttpError as e:
        error(str(e))
        sys.exit(1)


@docker.command()
@click.pass_context
def ping(ctx, **kwargs):
    client = ctx.obj['client']
    if client.ping():
        click.echo("Success")


@docker.command()
@options(_global_options)
@click.argument("image", type=Image, metavar='IMAGE')
@make_option("--no-input", is_flag=True)
@click.pass_context
@handle_lazo_error
def rm(ctx, image, no_input, **kwargs):
    client = ctx.obj['client']
    if image.tag:
        if client.exists(image):
            msg = """THIS IS A UNRECOVERABLE ACTION.

Continuing will irremediably remove image '%s' from the server.
Be sure to have a local copy.

Do you want to continue? """

            if no_input or click.confirm(msg % image.id):
                token = client.login()
                response = client.delete(f'/repositories/{image.account}/{image.image}/tags/{image.tag}/',
                                         raw=True,
                                         headers={'Authorization': f'JWT {token}'})
                if response.status_code == 204:
                    success("Deleted")
            else:
                error("Aborted")
        else:
            error("Tag does not exists")
    else:
        error(f"Cannot find {image.id}")



from click import BadParameter, UsageError

class LazoError(Exception):
    pass

class ServerConnectionError(LazoError):
    def __init__(self, url, exc=None):
        self.url = url
        self.reason = exc

    def __str__(self):
        return f"{self.url}. {str(self.reason or '')}"


class ServerSSLError(ServerConnectionError):
    def __str__(self):
        return 'Certificate verify failed. Try to use --insecure'


class HttpError(LazoError):
    def __init__(self, url, response, exc=None):
        self.url = url
        self.response = response
        self.exc = exc

    def __str__(self):
        return f"{self.response.status_code}: {self.url}. {str(self.exc or '')}"


class Http404(HttpError):
    pass


class InvalidName(UsageError):
    def __init__(self, name):
        self.name = name
        self.message = name


class InvalidCredentials(HttpError):
    def __str__(self):
        return "Invalid credential"


class RequiredParameter(UsageError):
    pass


class ExBadParameter(BadParameter):
    def format_message(self):
        return self.message


class ObjectNotFound(Exception):
    pass



from urllib.parse import urlparse

import click
from click.types import BoolParamType
from requests.auth import HTTPBasicAuth

from lazo.exceptions import ExBadParameter
from lazo.objects import DockerImage, RancherWorkload


class ExParamType(click.ParamType):

    def rfail(self, message, param=None, ctx=None):
        # message = f"Invalid value for {param}"
        raise ExBadParameter(message, ctx=ctx, param=param)


class UrlParamType(ExParamType):
    def convert(self, value, param, ctx):
        try:
            o = urlparse(value)
            assert not o.path.endswith('/')
        except AssertionError:
            self.fail(f"Url should not ends with '/'")
        except Exception:
            self.fail(f"Invalid url. Should be something like 'https://rancher.example.com:9000/v3/'")

        return value


Url = UrlParamType()


class VerbosityParamType(ExParamType):
    name = 'verbosity'

    def __init__(self) -> None:
        self.quit = False

    def convert(self, value, param, ctx):
        self.total = value
        if param.name == 'quit':
            self.quit = True
        if self.quit:
            value = -1
        return int(value)


Verbosity = VerbosityParamType()

ImagePullPolices = {'IfNotPresent': 'IfNotPresent',
                    'Always': 'Always',
                    'Never': 'Never',
                    'ifnotpresent': 'IfNotPresent',
                    'always': 'Always',
                    'never': 'Never',
                    'i': 'IfNotPresent',
                    'a': 'Always',
                    'n': 'Never',
                    }


class IChoice(click.Choice):

    def convert(self, value, param, ctx):
        if value in ImagePullPolices:
            value = ImagePullPolices[value]
        return super().convert(value, param, ctx)


# class TargetParamType(ExParamType):
#     def convert(self, value, param, ctx):
#         try:
#             parts = value.split(":")
#             assert len(parts) == 2
#         except Exception:
#             self.rfail(f"Invalid value '{value}' for TARGET. Please indicate target in the form 'namespace:workload' ")
#         return parts
#
#
# Target = TargetParamType()


class WorkloadParamType(ExParamType):
    def convert(self, value, param, ctx):
        try:
            return RancherWorkload(value)
        except Exception:
            self.rfail(f"Invalid  workload name '{value}'. "
                       f"Please use the form '[deployment:]namespace:workload' ")


Workload = WorkloadParamType()


class ProjectParamType(ExParamType):
    def convert(self, value, param, ctx):
        parts = [None, value]
        try:
            if ':' in value:
                parts = value.split(":")
                assert len(parts) == 2
        except Exception:
            self.rfail(
                f"Invalid value '{value}' for PROJECT. Please indicate project in the form 'clusterId:projectID' ")
        return parts


Project = ProjectParamType()


class AuthParamType(ExParamType):
    def convert(self, value, param, ctx):
        try:
            parts = value.split(":")
            assert len(parts) == 2
        except Exception:
            self.fail("Please indicate credentials as 'key:secret' ")
        return HTTPBasicAuth(*parts)


Auth = AuthParamType()


class ImageParamType(ExParamType):
    def convert(self, value, param, ctx):
        try:
            return DockerImage(value)
        except Exception as e:
            self.fail(f"{e}. Invalid '{value}' Please indicate image in the form 'account[/image[:tag]]' ")


Image = ImageParamType()


class DebugModeType(BoolParamType):

    def __call__(self, value, param=None, ctx=None):
        ctx.find_root().command.debug = True
        return True


DebugMode = DebugModeType()


class StdinAuthType(BoolParamType):

    def __call__(self, value, param=None, ctx=None):
        stdin = click.get_text_stream('stdin')
        credentials = stdin.read()
        return credentials.split(":")


StdinAuth = StdinAuthType()



#!/usr/bin/env python
import os
import sys
import warnings

import click
from urllib3.exceptions import InsecureRequestWarning

import lazo

from .params import _global_options, all_envs, make_option, options
from .utils import import_by_name

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])
warnings.simplefilter("ignore", InsecureRequestWarning)


class MyCLI(click.Group):

    def __init__(self, name=None, commands=None, **attrs):
        self.debug = False
        super().__init__(name, commands, **attrs)


def display(value):
    if value is None:
        return ''
    return str(value)


PROTECTED = ['auth']


@click.group(context_settings=CONTEXT_SETTINGS, cls=MyCLI, invoke_without_command=True, no_args_is_help=True)
@click.version_option(lazo.__version__)
@make_option('--env', is_flag=True, is_eager=True)
@make_option('--defaults', is_flag=True, is_eager=True)
@options(_global_options)
@click.pass_context
def cli(ctx, env, defaults, **kwargs):
    def get_value(opt):
        v = os.environ.get(opt.envvar, opt.default)
        if opt.is_flag:
            value = opt.type.convert(v, opt.name, click.get_current_context())
        else:
            value = os.environ.get(opt.envvar, opt.default) or ''
        if opt.name in PROTECTED:
            value = '***' if value else ''
        return str(value), 'from env' if opt.envvar in os.environ else ''

    if defaults:
        click.echo(f"{'Env':<20} {'Value':<50} Origin")
        for i, opt in sorted(all_envs.items()):
            name = opt.name
            value, descr = get_value(opt)
            click.echo(f"{name:<20} {value:<50} {descr}")
        sys.exit(0)
    elif env:
        click.echo(f"{'Env':<20} Value")
        for i, opt in sorted(all_envs.items()):
            value, in_env = get_value(opt)
            if in_env:
                click.echo(f"{i:<20} {value:<30}")
            else:
                click.echo(f"{i:<20} -- not set --")
        sys.exit(0)
    ctx.obj = {}


cli.add_command(import_by_name('lazo.commands.rancher.rancher'))
cli.add_command(import_by_name('lazo.commands.docker.docker'))

if __name__ == '__main__':
    cli()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import ast
import codecs
import os
import re

from setuptools import setup, find_packages

ROOT = os.path.realpath(os.path.join(os.path.dirname(__file__)))
script = os.path.join(ROOT, 'src', 'lazo', '__init__.py')

with open(script, 'rb') as f:
    content = f.read().decode('utf-8')
    _version_re = re.compile(r'__version__\s+=\s+(.*)')
    version = str(ast.literal_eval(_version_re.search(content).group(1)))

test_requires = ['pytest',
                 'pytest-cov',
                 'vcrpy'
                 ]
setup(name='lazo',
      version=version,
      description="""small utility to iteract with Rancher API""",
      long_description=codecs.open('README.md').read(),
      long_description_content_type='text/markdown',
      author='Stefano Apostolico',
      author_email='s.apostolico@gmail.com',
      url='https://github.com/saxix/lazo',
      package_dir={'': 'src'},
      packages=find_packages('src'),
      install_requires=['click==7.0',
                        'colorama==0.4.1',
                        'pygments',
                        'requests>=2.21.0',
                        'websocket-client==0.55.0',
                        ],
      test_requires=test_requires,
      extras_require={
          'test': test_requires,
      },
      entry_points={
          'console_scripts': [
              'lazo = lazo.__cli__:cli',
          ],
      },
      license='MIT',
      zip_safe=False,
      keywords='rancher',
      classifiers=[
          'Development Status :: 3 - Alpha',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python :: 3.6'
      ])



import inspect
import itertools
from functools import partial

import click
from click import Option
from click.decorators import _param_memo

from .types import Auth, DebugMode, IChoice, Project, StdinAuth, Url, Verbosity


class OOption(Option):
    def __init__(self, param_decls=None, show_default=True, prompt=False, confirmation_prompt=False, hide_input=False,
                 is_flag=None, flag_value=None, multiple=False, count=False, allow_from_autoenv=True, type=None,
                 help=None, hidden=False, show_choices=True, show_envvar=True, **attrs):
        super().__init__(param_decls, show_default, prompt, confirmation_prompt, hide_input, is_flag, flag_value,
                         multiple, count, allow_from_autoenv, type, help, hidden, show_choices, show_envvar, **attrs)


all_envs = {}


def make_option(*param_decls, **attrs):
    def decorator(f):
        # Issue 926, copy attrs, so pre-defined options can re-use the same cls=
        option_attrs = attrs.copy()

        if 'help' in option_attrs:
            option_attrs['help'] = inspect.cleandoc(option_attrs['help'])
        OptionClass = option_attrs.pop('cls', Option)
        opt = OptionClass(param_decls, **option_attrs)
        _param_memo(f, opt)
        if 'envvar' in attrs:
            e = attrs['envvar']
            all_envs[e] = opt
        return f

    return decorator


_global_options = [make_option('-v', '--verbosity',
                               default=1,
                               type=Verbosity,
                               help="verbosity level",
                               count=True),
                   make_option('-q', '--quiet',
                               help="no output",
                               default=0, is_flag=True, type=Verbosity),
                   make_option('-d',
                               '--dry-run',
                               is_flag=True,
                               help='dry-run mode'),
                   make_option('--debug',
                               default=False,
                               is_flag=True,
                               flag_value=True,
                               type=DebugMode,
                               help='debug mode'),
                   ]

_rancher_options = [make_option('-b',
                                '--base-url',
                                required=True,
                                type=Url,
                                envvar='RANCHER_BASE_URL',
                                cls=OOption,
                                help='Rancher base url.',
                                metavar='URL'),

                    make_option('--auth',
                                envvar='RANCHER_AUTH',
                                help='Rancher API key:secret',
                                type=Auth,
                                cls=OOption,
                                default=None,
                                metavar='TEXT'),
                    make_option('--stdin',
                                is_flag=True,
                                help='Read credentials from stdin'),
                    make_option('-i',
                                '--insecure',
                                is_flag=True,
                                envvar='RANCHER_INSECURE',
                                cls=OOption,
                                help='Ignore verifying the SSL certificate'),
                    make_option('-n',
                                '--use-names',
                                envvar='RANCHER_USE_NAMES',
                                is_flag=True,
                                help='Use target names instead of Rancher Id(s)'),
                    ]
CLUSTER = make_option('-c',
                      '--cluster',
                      required=True,
                      envvar='RANCHER_CLUSTER',
                      help='Rancher cluster key.',
                      cls=OOption,
                      metavar='TEXT')
PROJECT = make_option('-p',
                      '--project',
                      required=True,
                      type=Project,
                      envvar='RANCHER_PROJECT',
                      cls=OOption,
                      help='Rancher project key',
                      metavar='PROJECT')
# WORKLOAD = make_option('-w',
#                        '--workload',
#                        type=Workload,
#                        envvar='RANCHER_WORKLOAD',
#                        cls=OOption,
#                        help='Rancher workload.',
#                        metavar='TEXT')

_workload_options = [CLUSTER,
                     PROJECT,
                     make_option('--pull', 'pull_policy',
                                 type=IChoice(['IfNotPresent', 'Always', 'Never'], False),
                                 default='Always',
                                 help='Rancher ImagePullPolicy'),
                     make_option('--name',
                                 help='Workload new name'),
                     ]

_docker_options = [make_option('-r',
                               '--repository',
                               default='https://hub.docker.com/v2',
                               envvar='DOCKER_REPOSITORY',
                               metavar='URL',
                               cls=OOption,
                               help='Docker repository'),
                   make_option('-u', '--username',
                               help='username'),
                   make_option('-p', '--password',
                               help='password'),
                   make_option('--check-image/--no-check-image',
                               is_flag=True,
                               default=True,
                               help='Do not check Docker repository'),
                   make_option('--stdin',
                               type=StdinAuth,
                               is_flag=True,
                               help='Read credentials from stdin'),
                   ]


def options(*opts):
    def decorator(func):
        for option in reversed(list(itertools.chain(*opts))):
            func = option(func)
        return func

    return decorator


class Config(object):

    def __init__(self, root):
        self.root = root
        self.client = None
        self.storage = {}

    def __repr__(self):
        return '<Config>'


pass_config = click.make_pass_decorator(Config)



import importlib
import json

from pygments import highlight
from pygments.formatters.terminal import TerminalFormatter
from pygments.lexers.data import JsonLexer


def sizeof(num, suffix='B'):
    for unit in ['', 'Ki', 'Mi', 'Gi', 'Ti', 'Pi', 'Ei', 'Zi']:
        if abs(num) < 1024.0:
            return "%3.1f%s%s" % (num, unit, suffix)
        num /= 1024.0
    return "%.1f%s%s" % (num, 'Yi', suffix)


def import_by_name(name):
    """dynamically load a class from a string

    es:
        klass = import_by_name('my_package.my_module.my_class')
        some_object = klass()

    :param name:
    :return:

    """
    if '.' not in name:
        raise ValueError("Cannot import '{}'".format(name))
    class_data = name.split(".")
    module_path = ".".join(class_data[:-1])
    class_str = class_data[-1]
    module = importlib.import_module(module_path)
    try:
        return getattr(module, class_str)
    except AttributeError:
        raise AttributeError('Unable to import {}. '
                             '{} does not have {} attribute'.format(name,
                                                                    module,
                                                                    class_str))


def jprint(obj, colors=True):
    formatted_json = json.dumps(obj, sort_keys=True, indent=4)
    if colors:
        colorful_json = highlight(formatted_json, JsonLexer(), TerminalFormatter())
        print(colorful_json)
    else:
        print(formatted_json)


def prepare_command(cmds):
    if not isinstance(cmds, (list, tuple)):
        cmds = cmds.split()
    return list(zip(['command'] * len(cmds), cmds))



import sys
from functools import partial

import click


def secho(*args, fg='white'):
    click.secho(" ".join(map(str, args)), fg=fg)


def fail(*args):
    secho(*args, fg='red')
    sys.exit(1)


echo = partial(secho, fg='white')
success = partial(secho, fg='green')
warn = partial(secho, fg='orange')
error = partial(secho, fg='red')
# click.echo(click.style('Hello World!', fg='green'))



import sys
import urllib

import click
from click import argument, UsageError

from ..__cli__ import cli
from ..clients import DockerClient, RancherClient, handle_lazo_error
from ..out import echo, error, success
from ..params import (CLUSTER, PROJECT, _docker_options, _global_options,
                      _rancher_options, make_option, options, )
from ..types import DockerImage, Image, RancherWorkload, Workload
from ..utils import jprint, prepare_command


@cli.group()
@options(_global_options, _rancher_options)
@click.pass_context
def rancher(ctx, base_url, insecure, auth, use_names, debug, **kwargs):
    ctx.obj['client'] = RancherClient(base_url, auth=auth,
                                      verify=not insecure,
                                      use_names=use_names,
                                      debug=ctx.find_root().command.debug)


@rancher.command()
@click.pass_context
@handle_lazo_error
def ping(ctx, **kwargs):
    client = ctx.obj['client']
    if client.ping():
        success("Success")
    else:
        error("Fail")


@rancher.command()
@click.pass_context
@handle_lazo_error
def settings(ctx, **kwargs):
    client = ctx.obj['client']
    ret = client.get('/settings')
    jprint(ret)


@rancher.command()
@click.pass_context
@handle_lazo_error
def login(ctx, **kwargs):
    client = ctx.obj['client']
    if client.ping():
        success(f"Successfully logged in to {client.base_url}")
    else:
        error("Fail")


@rancher.command()
@options(_global_options)
@options([CLUSTER, PROJECT])
@click.option('-w',
              '--workload',
              type=Workload,
              help='Rancher workload.',
              metavar='TEXT')
@click.pass_context
@handle_lazo_error
def info(ctx, cluster, project, workload: RancherWorkload, verbosity, **kwargs):
    client = ctx.obj['client']

    client.cluster = cluster
    if project:
        client.project = project

    if workload:
        echo('Workload infos:')
        info = client.get_workload(workload.id)
        if 'containers' in info:
            for e in info['containers']:
                echo("Image:", e['image'])
                echo("imagePullPolicy:", e['imagePullPolicy'])
        if 'publicEndpoints' in info:
            for e in info['publicEndpoints']:
                echo("Ingress:", e['ingressId'])
                echo("Hostname:", e['hostname'])
    elif project:
        echo('Project workloads:')
        for workload in client.list_workloads():
            echo(f"\t{workload[0]:>20}    {workload[1]:<40}")
    elif cluster:
        response = client.get(f'/clusters/{client.cluster}/projects')
        echo(f'Projects on cluster: {client.cluster}')
        for project in response['data']:
            echo(f"\t{project['name']:>15}    {project['id']:<40}")
    else:
        echo('Clusters:')
        for entry in client.list_clusters():
            echo(f"\t{entry[0]:>20}   {entry[1]:<40}")


@rancher.command()
@options([CLUSTER, PROJECT])
@click.pass_context
def pods(ctx, cluster, project):
    client = ctx.obj['client']
    client.cluster = cluster
    client.project = project
    ret = client.list_pods()
    for w in ret:
        echo(f'{w["workloadId"]:<50} {w["id"]}')


@rancher.command()
@options([CLUSTER, PROJECT])
@click.pass_context
def containers(ctx, cluster, project):
    client = ctx.obj['client']
    client.cluster = cluster
    client.project = project
    ret = client.list_containers()
    for w in ret:
        # TODO: remove me
        print(111, "rancher.py:120", w)
        # echo(f'{w["workloadId"]:<50} {w["id"]}')
        # for c in w["containers"]:
        #     echo(f'\t{c["id"]}')
    # jprint(ret)


@rancher.command()
@options(_global_options, _docker_options)
@options([CLUSTER, PROJECT])
@make_option('--workload', '-w', 'workloads', type=Workload, metavar='WORKLOAD', required=True, multiple=True)
@make_option('--image', '-i', type=Image, metavar='IMAGE', required=True, )
@make_option('--check/--no-check', is_flag=True, default=True)
@click.pass_context
@handle_lazo_error
def upgrade(ctx, cluster, project, workloads: RancherWorkload, image: DockerImage, check,
            repository, username, password, **kwargs):
    client = ctx.obj['client']
    if check:
        docker = DockerClient(repository, username=username, password=password)
        if not docker.exists(image):
            error(f"Cannot find image '{image.id}' on {repository}")
            error(f"Run lazo docker info '{image.id}' to check available tags")
            sys.exit(1)
    client.cluster = cluster
    client.project = project
    # namespace, workload_name = workload

    for workload in workloads:
        echo(f"Upgrading workload '{workload.id}' on project '{client.cluster}:{client.project}' to '{image.id}'")

        client.upgrade(workload, image)
        info = client.get_workload(workload)
        if 'containers' in info:
            for e in info['containers']:
                echo("Image:", e['image'])
        if 'publicEndpoints' in info:
            for ep in info['publicEndpoints']:
                echo("Ingress:", ep['ingressId'])
                echo("Hostname:", ep.get('hostname', ''))


@rancher.command()
@options(_global_options, _docker_options)
@options([CLUSTER, PROJECT])
@argument('workload', type=Workload, metavar='NAME')
@argument('command', nargs=-1)
# @make_option('-c', '--check', is_flag=True)
# @make_option('-c', '--dry-run', is_flag=True)
@click.pass_context
@handle_lazo_error
def shell(ctx, cluster, project, workload: RancherWorkload, command, **kwargs):
    client = ctx.obj['client']
    client.cluster = cluster
    client.project = project
    pod = client.get_pod(workload)
    cmds = [('container', pod.workload.name),
            ('stdin', '1'),
            ('stdout', '1'),
            ('stderr', '1'),
            ('tty', '1'),
            ]
    cmds.extend(prepare_command(command))
    qs = urllib.parse.urlencode(cmds)
    try:
        url = f'/k8s/clusters/{client.cluster}/api/v1/namespaces/{workload.namespace}/pods/{pod.name}/exec?{qs}'
        ret = client.ws(url)
    except Exception as e:
        error(e)
        sys.exit(1)

    print(ret)
